#include "sacd_dsdiff.h"

sacd_dsdiff_t::sacd_dsdiff_t() {
}

sacd_dsdiff_t::~sacd_dsdiff_t() {
}

int sacd_dsdiff_t::get_channels() {
	return m_channel_count;
}

int sacd_dsdiff_t::get_loudspeaker_config() {
	return m_loudspeaker_config;
}

int sacd_dsdiff_t::get_samplerate() {
	return m_samplerate;
}

uint64_t sacd_dsdiff_t::get_size() {
	return m_data_size;
}

uint64_t sacd_dsdiff_t::get_offset() {
	return m_file->get_position(m_abort) - m_data_offset;
}

double sacd_dsdiff_t::get_duration() {
	return m_frame_count / 75.0;
}

void sacd_dsdiff_t::get_info(t_uint32 p_subsong, file_info& p_info) {
	if (m_id3_size > 0) {
		uint64_t saved_offset = m_file->get_position(m_abort);
		try {
			service_ptr_t<reader_limited> r = new service_impl_t<reader_limited>();
			r->init(m_file, m_id3_offset, m_id3_offset + m_id3_size, m_abort);
			tag_processor::read_id3v2(r, p_info, m_abort);
		}
		catch (exception_io_data) {
		}
		m_file->seek_ex(saved_offset, file::seek_from_beginning, m_abort);
	}
}

bool sacd_dsdiff_t::is_dst() {
	return m_dst_encoded != 0;
}

bool sacd_dsdiff_t::open(service_ptr_t<file> p_file) {
	m_file = p_file;
	m_id3_size = 0;
	Chunk ck;
	ID id;
	if (!(m_file->read(&ck, sizeof(ck), m_abort) == sizeof(ck) && ck.has_id("FRM8"))) {
		return false;
	}
	if (!(m_file->read(&id, sizeof(id), m_abort) == sizeof(id) && id.has_id("DSD "))) {
		return false;
	}
	for (;;) {
		if (!(m_file->read(&ck, sizeof(ck), m_abort) == sizeof(ck))) {
			return false;
		}
		if (ck.has_id("FVER") && ck.get_size() == 4) {
			uint32_t version;
			if (!(m_file->read(&version, sizeof(version), m_abort) == sizeof(version))) {
				return false;
			}
			m_version = hton32(version);
		}
		else if (ck.has_id("PROP")) {
			if (!(m_file->read(&id, sizeof(id), m_abort) == sizeof(id) && id.has_id("SND "))) {
				return false;
			}
			uint64_t id_prop_size = ck.get_size() - sizeof(id);
			uint64_t id_prop_read = 0;
			while (id_prop_read < id_prop_size) {
				if (!(m_file->read(&ck, sizeof(ck), m_abort) == sizeof(ck))) {
					return false;
				}
				if (ck.has_id("FS  ") && ck.get_size() == 4) {
					uint32_t samplerate;
					if (!(m_file->read(&samplerate, sizeof(samplerate), m_abort) == sizeof(samplerate))) {
						return false;
					}
					m_samplerate = hton32(samplerate);
				}
				else if (ck.has_id("CHNL")) {
					uint16_t channel_count;
					if (!(m_file->read(&channel_count, sizeof(channel_count), m_abort) == sizeof(channel_count))) {
						return false;
					}
					m_channel_count = hton16(channel_count);
					switch (m_channel_count) {
					case 2:
						m_loudspeaker_config = 0; 
						break;
					case 5:
						m_loudspeaker_config = 3; 
						break;
					case 6:
						m_loudspeaker_config = 4; 
						break;
					default:
						m_loudspeaker_config = 65535; 
						break;
					}
					m_file->skip(ck.get_size() - sizeof(channel_count), m_abort);
				}
				else if (ck.has_id("CMPR")) {
					if (!(m_file->read(&id, sizeof(id), m_abort) == sizeof(id))) {
						return false;
					}
					if (id.has_id("DSD ")) {
						m_dst_encoded = 0;
					}
					if (id.has_id("DST ")) {
						m_dst_encoded = 1;
					}
					m_file->skip(ck.get_size() - sizeof(id), m_abort);
				}
				else if (ck.has_id("LSCO")) {
					uint16_t loudspeaker_config;
					if (!(m_file->read(&loudspeaker_config, sizeof(loudspeaker_config), m_abort) == sizeof(loudspeaker_config))) {
						return false;
					}
					m_loudspeaker_config = hton16(loudspeaker_config);
					m_file->skip(ck.get_size() - sizeof(loudspeaker_config), m_abort);
				}
				else if (ck.has_id("ID3 ")) {
					m_id3_offset = m_file->get_position(m_abort);
					m_id3_size = ck.get_size();
					m_file->skip(ck.get_size(), m_abort);
				}
				else {
					m_file->skip(ck.get_size(), m_abort);
				}
				id_prop_read += sizeof(ck) + ck.get_size();
			}
			continue;
		}
		else if (ck.has_id("DSD ")) {
			m_data_offset = m_file->get_position(m_abort);
			m_data_size = ck.get_size();
			m_framerate = 75;
			m_frame_size = m_samplerate / 8 * m_channel_count / m_framerate;
			m_frame_count = (uint32_t)(m_data_size / m_frame_size);
			return true;
		}
		else if (ck.has_id("DST ")) {
			m_data_offset = m_file->get_position(m_abort);
			m_data_size = ck.get_size();
			if (!(m_file->read(&ck, sizeof(ck), m_abort) == sizeof(ck) && ck.has_id("FRTE") && ck.get_size() == 6)) {
				return false;
			}
			m_data_size -= sizeof(ck) + ck.get_size();
			uint32_t frame_count;
			if (!(m_file->read(&frame_count, sizeof(frame_count), m_abort) == sizeof(frame_count))) {
				return false;
			}
			m_frame_count = hton32(frame_count);
			uint16_t framerate;
			if (!(m_file->read(&framerate, sizeof(framerate), m_abort) == sizeof(framerate))) {
				return false;
			}
			m_framerate = hton16(framerate);
			m_frame_size = m_samplerate / 8 * m_channel_count / m_framerate;
			return true;
		}
		else {
			m_file->skip(ck.get_size(), m_abort);
		}
	}
	return false;
}

bool sacd_dsdiff_t::close() {
	return true;
}

bool sacd_dsdiff_t::read_frame(uint8_t* frame_data, size_t* frame_size) {
	if (m_dst_encoded) {
		Chunk ck;
		while (m_file->read(&ck, sizeof(ck), m_abort) == sizeof(ck)) {
			if (ck.has_id("DSTF")) {
				if (m_file->read(frame_data, (t_size)ck.get_size(), m_abort) == ck.get_size()) {
					*frame_size = (size_t)ck.get_size();
					m_file->skip(ck.get_size() & 1, m_abort);
					return true;
				}
				break;
			}
			else if (ck.has_id("DSTC")) {
				if (m_file->read(&ck, sizeof(ck), m_abort) == sizeof(ck)) {
					m_file->skip(ck.get_size(), m_abort);
					m_file->skip(ck.get_size() & 1, m_abort);
				}
			}
			else {
				m_file->seek_ex(1 - (int)sizeof(ck), file::seek_from_current, m_abort);
			}
		}
	}
	else {
		uint64_t position = m_file->get_position(m_abort);
		*frame_size = min(m_frame_size, (uint32_t)max(0, (int64_t)(m_data_offset + m_data_size) - (int64_t)position));
		if (frame_size > 0) {
			*frame_size = m_file->read(frame_data, *frame_size, m_abort);
			*frame_size -= *frame_size % m_channel_count;
			if (*frame_size > 0) {
				return true;
			}
		}
	}
	return false;
}

bool sacd_dsdiff_t::seek(uint64_t offset) {
	if (offset < m_data_size) {
		if (!m_dst_encoded) {
			offset -= offset % m_frame_size;
		}
		m_file->seek(m_data_offset + offset, m_abort);
		return true;
	}
	return false;
}
